(*
    HLins: insert http-links into HTML documents.
    See http://www.lri.fr/~treinen/hlins

    Copyright (C) 1999 Ralf Treinen <treinen@lri.fr>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*)


open Arg;;
open Str;;
open Read_databases;;
open Automaton;;
open Build_automaton;;
open Run_automaton;;
open Errors;;
open Read_html;;

let version = "0.31";;


(*************************************************************************)
(* process options *)

let dblist = ref [];;
let out_filename = ref "";;
let in_filename = ref "";;
let quiet = ref false;;

parse
  [ ( "-db",
      String (function s -> dblist := split (regexp "[ \t\n]+") s),
      "list of address database files"
    );
    ( "-o",
      String (function s -> out_filename := s),
      "output file"
    );
    ( "-quiet",
      Set quiet,
      "supress diagnostic output"
    )
  ]
  (function s -> in_filename := s)
  "Hlins: insert urls into a file."
;;

if !out_filename = !in_filename && !out_filename <> ""
then stopwitherror "Input and output files must differ";;


(***************************************************************************)
(* now do it ! *)



(* (allequal l), where l is a list, tests whether all elements
   of l are equal
*)
let allequal = function
    [] -> true
  | (h::l) -> List.for_all (function x -> x=h) l;;



(* (geturl names urls foundinds expandinds) returns a url for name
   where name is the array of names, urls the array of urls, foundinds
   the set of indices of the word found and expandinds the set of
   indices of words that the word found is an abbreviation of.
*) 
let geturl names urls foundinds expandinds s =
  let foundurls = List.map (Array.get urls) (foundinds@expandinds)
  in begin
    if !quiet || allequal foundurls
    then List.hd foundurls
    else 
      begin
	prerr_string ("A conflict occured on the name "^s^":\n");
	List.iter (function i ->
		     prerr_string ("  has url "^urls.(i)^"\n")) foundinds;
	List.iter (function i ->
		     prerr_string ("  abbreviates "^names.(i)^
				   " which has url "^
				   urls.(i)^"\n")) expandinds; 
	prerr_endline ("Choosing " ^ (List.hd  foundurls));
	List.hd foundurls
      end;
    end
;;

let hitcount = ref 0;;

let replace_function names urls foundinds expandinds s= 
  incr hitcount;
  "<a href=\""^
  (geturl names urls foundinds expandinds s^
   "\">" ^ s ^ "</a>")
;;


let doit inc ouc =
  if not !quiet then
      prerr_string ("Hlins version "^version^".\n");
  let (name_list,url_list) = read_dbs !dblist in
  let names = Array.of_list name_list
  and urls = Array.of_list url_list
  (*in    print_automaton (build_automaton names) *)
  in run
       (build_automaton names)
       (Lexing.from_channel inc)
       next_char
       (replace_function names urls)
       ouc;
    if not !quiet
    then prerr_string
      ("Number of insertions: " ^ (string_of_int !hitcount) ^"\n");
    prerr_newline();
    close_in inc;
    close_out ouc
;;


Printexc.catch doit
  (if !in_filename <> ""
   then
     try open_in !in_filename
     with Sys_error s -> stopwitherror ("Cannot read  input file "^s)
   else
     stdin
  )
  (if !out_filename <> ""
   then
     try open_out !out_filename
     with Sys_error s -> stopwitherror ("Cannot open output file "^s)
   else
     stdout
  )
;;
















